import dataclasses
import time
from enum import IntEnum
from typing import Optional
import base64

import pbjson
from cryptography.fernet import Fernet


class Phase(IntEnum):
    OPEN = 1
    DATA = 2
    CLOSE = 3


TOKENS = {
    'phase': 'pe',
    'ack': 'ak',
    'token': 'tn',
    'host': 'ht',
    'port': 'pt',
    'seq': 'sq',
    'data': 'da',
    'drain': 'dn',
}


@dataclasses.dataclass(kw_only=True)
class Packet:
    phase: Phase
    ack: bool
    token: Optional[str] = None
    host: Optional[str] = None
    port: Optional[int] = None
    seq: Optional[int] = None
    data: Optional[bytes] = None
    drain: Optional[bool] = None

    @staticmethod
    def packet(*, data: bytes, fernet: Fernet) -> 'Packet':
        json = pbjson.loads(fernet.decrypt(base64.urlsafe_b64encode(data)))
        return Packet(
            phase=Phase(json.get(TOKENS['phase'])),
            ack=json.get(TOKENS['ack']),
            token=json.get(TOKENS['token'], None),
            host=json.get(TOKENS['host'], None),
            port=json.get(TOKENS['port'], None),
            seq=json.get(TOKENS['seq'], None),
            data=json.get(TOKENS['data'], None),
            drain=json.get(TOKENS['drain'], None),
        )

    def binary(self, *, fernet: Fernet) -> bytes:
        json = {TOKENS[k]: v for k, v in self.__dict__.items() if v is not None}
        return base64.urlsafe_b64decode(fernet.encrypt(pbjson.dumps(json)))


@dataclasses.dataclass(kw_only=True)
class Sent:
    data: bytes
    timestamp: int = int(time.time())
    retries: int = 1


@dataclasses.dataclass(kw_only=True)
class Received:
    data: bytes
    drain: bool



import pytest

from ouija import Phase, Packet


@pytest.mark.parametrize('packet', (
    Packet(phase=Phase.OPEN, ack=False, token='secret', host='example.com', port=443),
    Packet(phase=Phase.OPEN, ack=True, token='secret'),
    Packet(phase=Phase.DATA, ack=False, seq=0, data=b'test data 1', drain=False),
    Packet(phase=Phase.DATA, ack=False, seq=1, data=b'test data 2', drain=True),
    Packet(phase=Phase.DATA, ack=True, seq=0),
    Packet(phase=Phase.DATA, ack=True, seq=1),
    Packet(phase=Phase.CLOSE, ack=False),
    Packet(phase=Phase.CLOSE, ack=True),
))
def test_packet(packet, fernet_test):
    encoded = packet.binary(fernet=fernet_test)
    decoded = Packet.packet(data=encoded, fernet=fernet_test)
    assert decoded == packet



import asyncio
import logging
import os
from typing import Dict

from .tuning import Tuning
from .relay import Relay
from .telemetry import Telemetry
from .rawparser import RawParser


logging.basicConfig(
    format='%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d:%H:%M:%S',
    level=logging.ERROR,
)
logger = logging.getLogger(__name__)


class Interface:
    telemetry: Telemetry
    tuning: Tuning
    proxy_host: str
    proxy_port: int
    index: int
    sessions: Dict[int, asyncio.Task]

    def __init__(self, *, telemetry: Telemetry, tuning: Tuning, proxy_host: str, proxy_port: int) -> None:
        self.telemetry = telemetry
        self.tuning = tuning
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.index = 0
        self.sessions = dict()

    async def https_handler(
            self,
            *,
            reader: asyncio.StreamReader,
            writer: asyncio.StreamWriter,
            remote_host: str,
            remote_port: int,
    ) -> None:
        relay = Relay(
            telemetry=self.telemetry,
            tuning=self.tuning,
            reader=reader,
            writer=writer,
            proxy_host=self.proxy_host,
            proxy_port=self.proxy_port,
            remote_host=remote_host,
            remote_port=remote_port,
        )
        await relay.serve()

    async def handle_session(self, *, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        data = await reader.readuntil(b'\r\n\r\n')
        request = RawParser(data=data)
        if request.error:
            logger.error('Parse error')
        elif request.method == 'CONNECT':
            await self.https_handler(
                reader=reader,
                writer=writer,
                remote_host=request.host,
                remote_port=request.port,
            )
        else:
            logger.error(f'{request.method} method is not supported')

    async def session(self, *, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        try:
            await asyncio.wait_for(self.handle_session(reader=reader, writer=writer), self.tuning.serving_timeout)
        except asyncio.TimeoutError:
            logger.error('Timeout')

    async def serve(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        self.sessions[self.index] = asyncio.create_task(self.session(reader=reader, writer=writer))
        self.index += 1

    async def cleanup(self) -> None:    # pragma: no cover
        while True:
            await asyncio.sleep(1)
            for index in sorted(self.sessions.keys()):
                if self.sessions[index].done():
                    self.sessions.pop(index)
            self.telemetry.link(links=len(self.sessions))

    async def monitor(self) -> None:    # pragma: no cover
        while True:
            await asyncio.sleep(1)
            os.system('clear')
            print(self.telemetry)



from unittest.mock import AsyncMock

import pytest
from pytest_mock import MockerFixture

from ouija import Packet, Phase


@pytest.mark.asyncio
async def test_link_sendto(link_test, data_test):
    await link_test.sendto(data=data_test)
    link_test.proxy.transport.sendto.assert_called()


@pytest.mark.asyncio
async def test_link_on_open(link_test, token_test, mocker: MockerFixture):
    async def open_connection(*args, **kwargs):
        return AsyncMock(), AsyncMock()

    mocked_asyncio = mocker.patch('ouija.link.asyncio')
    mocked_asyncio.open_connection = open_connection
    link_test.serve = AsyncMock()
    link_test.send_ack_open = AsyncMock()
    packet = Packet(
        phase=Phase.OPEN,
        ack=False,
        token=token_test,
        host='example.com',
        port=443,
    )
    assert await link_test.on_open(packet=packet)
    assert link_test.opened.is_set()
    link_test.serve.assert_called()
    link_test.send_ack_open.assert_awaited()


@pytest.mark.asyncio
async def test_link_on_open_opened(link_test, token_test):
    link_test.opened.set()
    link_test.send_ack_open = AsyncMock()
    packet = Packet(
        phase=Phase.OPEN,
        ack=False,
        token=token_test,
        host='example.com',
        port=443,
    )
    assert not await link_test.on_open(packet=packet)
    link_test.send_ack_open.assert_awaited()


@pytest.mark.asyncio
async def test_link_on_close(link_test, token_test):
    link_test.proxy.links = {link_test.addr: link_test}
    await link_test.on_close()
    assert not link_test.proxy.links



import datetime

from pytest_mock import MockerFixture


def test_telemetry_link(telemetry_test):
    telemetry_test.link(links=10)
    assert telemetry_test.links == 10


def test_telemetry_open(telemetry_test):
    telemetry_test.open()
    assert telemetry_test.opened == 1


def test_telemetry_close(telemetry_test):
    telemetry_test.close()
    assert telemetry_test.closed == 1


def test_telemetry_send(telemetry_test, data_test):
    telemetry_test.send(data=data_test)
    assert telemetry_test.packets_sent == 1
    assert telemetry_test.bytes_sent == 9
    assert telemetry_test.min_packet_size == 9
    assert telemetry_test.max_packet_size == 9


def test_telemetry_recv(telemetry_test, data_test):
    telemetry_test.recv(data=data_test)
    assert telemetry_test.packets_recv == 1
    assert telemetry_test.bytes_recv == 9
    assert telemetry_test.min_packet_size == 9
    assert telemetry_test.max_packet_size == 9


def test_telemetry_processing_error(telemetry_test):
    telemetry_test.processing_error()
    assert telemetry_test.processing_errors == 1


def test_telemetry_token_error(telemetry_test):
    telemetry_test.token_error()
    assert telemetry_test.token_errors == 1


def test_telemetry_type_error(telemetry_test):
    telemetry_test.type_error()
    assert telemetry_test.type_errors == 1


def test_telemetry_timeout_error(telemetry_test):
    telemetry_test.timeout_error()
    assert telemetry_test.timeout_errors == 1


def test_telemetry_connection_error(telemetry_test):
    telemetry_test.connection_error()
    assert telemetry_test.connection_errors == 1


def test_telemetry_serving_error(telemetry_test):
    telemetry_test.serving_error()
    assert telemetry_test.serving_errors == 1


def test_telemetry_resending_error(telemetry_test):
    telemetry_test.resending_error()
    assert telemetry_test.resending_errors == 1


def test_telemetry_send_buf_overload(telemetry_test):
    telemetry_test.send_buf_overload()
    assert telemetry_test.send_buf_overloads == 1


def test_telemetry_recv_buf_overload(telemetry_test):
    telemetry_test.recv_buf_overload()
    assert telemetry_test.recv_buf_overloads == 1


def test_telemetry(telemetry_test, mocker: MockerFixture):
    timestamp = datetime.datetime.now()
    mocked_datetime = mocker.patch('ouija.telemetry.datetime')
    mocked_datetime.datetime.now.return_value = timestamp
    expected = \
        f'{timestamp}\n' \
        f'\tlinks: 0\n' \
        f'\topened|closed: 0|0\n' \
        f'\tpackets sent|received: 0|0\n' \
        f'\tbytes sent|received: 0|0\n' \
        f'\tmin|max packet size: 0|0\n' \
        f'\tprocessing|token|type errors: 0|0' \
        f'|0\n' \
        f'\ttimeout|connection|serving|resending errors: 0|0' \
        f'|0|0\n' \
        f'\tsend|recv buf overloads: 0|0'
    assert str(telemetry_test) == expected



__author__ = 'Anton Smolin'
__copyright__ = 'Copyright (c) 2023 Anton Smolin'
__license__ = 'MIT'
__version__ = '0.3.0'


from .packet import Phase, Packet
from .telemetry import Telemetry
from .tuning import Tuning
from .ouija import Ouija
from .rawparser import RawParser
from .relay import Relay
from .interface import Interface
from .link import Link
from .proxy import Proxy



from unittest.mock import Mock, AsyncMock

import pytest


def test_proxy_connection_made(proxy_test):
    mock = Mock()
    proxy_test.connection_made(mock)
    assert proxy_test.transport == mock


@pytest.mark.asyncio
async def test_proxy_datagram_received_async(proxy_test, link_test, data_test):
    proxy_test.links = {('127.0.0.1', 60000): link_test}
    link_test.process = AsyncMock()
    await proxy_test.datagram_received_async(data=data_test, addr=('127.0.0.1', 60000))
    link_test.process.assert_awaited()


@pytest.mark.asyncio
async def test_proxy_datagram_received(proxy_test, data_test):
    proxy_test.datagram_received_async = AsyncMock()
    proxy_test.datagram_received(data_test, ('127.0.0.1', 60000))
    proxy_test.datagram_received_async.assert_called()


@pytest.mark.asyncio
async def test_proxy_connection_lost(proxy_test):
    proxy_test.transport = AsyncMock()
    proxy_test.connection_lost(Exception())
    proxy_test.transport.close.assert_called()



from unittest.mock import AsyncMock

import pytest
from pytest_mock import MockerFixture

from ouija import RawParser


@pytest.mark.asyncio
async def test_interface_https_handler(interface_test, relay_test, mocker: MockerFixture):
    mocked_relay = mocker.patch('ouija.interface.Relay')
    mocked_relay.return_value = relay_test
    relay_test.serve = AsyncMock()
    await interface_test.https_handler(
        reader=AsyncMock(),
        writer=AsyncMock(),
        remote_host='example.com',
        remote_port=443,
    )
    relay_test.serve.assert_awaited()


@pytest.mark.asyncio
async def test_interface_handle_session(interface_test):
    reader = AsyncMock()
    reader.readuntil = AsyncMock(return_value=b'CONNECT https://example.com:443 HTTP/1.1')
    interface_test.https_handler = AsyncMock()
    await interface_test.handle_session(reader=reader, writer=AsyncMock())
    interface_test.https_handler.assert_awaited()


@pytest.mark.asyncio
async def test_interface_handle_session_request_error(interface_test, mocker: MockerFixture):
    reader = AsyncMock()
    reader.readuntil = AsyncMock(return_value=b'CONNECT https://example.com:443 HTTP/1.1')
    interface_test.https_handler = AsyncMock()
    mocked_rawparser = mocker.patch('ouija.interface.RawParser')
    mocked_rawparser.return_value = RawParser(data=b'')
    await interface_test.handle_session(reader=reader, writer=AsyncMock())
    interface_test.https_handler.assert_not_awaited()


@pytest.mark.asyncio
async def test_interface_handle_session_method_error(interface_test):
    reader = AsyncMock()
    reader.readuntil = AsyncMock(return_value=b'GET example.com HTTP/1.1')
    interface_test.https_handler = AsyncMock()
    await interface_test.handle_session(reader=reader, writer=AsyncMock())
    interface_test.https_handler.assert_not_awaited()


@pytest.mark.asyncio
async def test_interface_session(interface_test):
    interface_test.handle_session = AsyncMock()
    await interface_test.session(reader=AsyncMock(), writer=AsyncMock())
    interface_test.handle_session.assert_awaited()


@pytest.mark.asyncio
@pytest.mark.xfail(raises=TimeoutError)
async def test_interface_session_timeouterror(interface_test):
    interface_test.handle_session = AsyncMock()
    interface_test.handle_session.side_effect = TimeoutError()
    await interface_test.session(reader=AsyncMock(), writer=AsyncMock())
    interface_test.handle_session.assert_awaited()


@pytest.mark.asyncio
async def test_interface_serve(interface_test):
    interface_test.session = AsyncMock()
    await interface_test.serve(reader=AsyncMock(), writer=AsyncMock())
    interface_test.session.assert_called()



import dataclasses
import datetime


@dataclasses.dataclass(kw_only=True)
class Telemetry:
    links: int = 0
    opened: int = 0
    closed: int = 0
    packets_sent: int = 0
    packets_recv: int = 0
    bytes_sent: int = 0
    bytes_recv: int = 0
    min_packet_size: int = 0
    max_packet_size: int = 0
    processing_errors: int = 0
    token_errors: int = 0
    type_errors: int = 0
    timeout_errors: int = 0
    connection_errors: int = 0
    serving_errors: int = 0
    resending_errors: int = 0
    send_buf_overloads: int = 0
    recv_buf_overloads: int = 0

    def __str__(self) -> str:
        return \
            f'{datetime.datetime.now()}\n' \
            f'\tlinks: {self.links:,}\n' \
            f'\topened|closed: {self.opened:,}|{self.closed:,}\n' \
            f'\tpackets sent|received: {self.packets_sent:,}|{self.packets_recv:,}\n' \
            f'\tbytes sent|received: {self.bytes_sent:,}|{self.bytes_recv:,}\n' \
            f'\tmin|max packet size: {self.min_packet_size}|{self.max_packet_size}\n' \
            f'\tprocessing|token|type errors: {self.processing_errors:,}|{self.token_errors:,}' \
            f'|{self.type_errors:,}\n' \
            f'\ttimeout|connection|serving|resending errors: {self.timeout_errors:,}|{self.connection_errors:,}' \
            f'|{self.serving_errors:,}|{self.resending_errors:,}\n' \
            f'\tsend|recv buf overloads: {self.send_buf_overloads:,}|{self.recv_buf_overloads:,}'

    def link(self, *, links: int) -> None:
        self.links = links

    def open(self) -> None:
        self.opened += 1

    def close(self) -> None:
        self.closed += 1

    def send(self, *, data: bytes) -> None:
        self.packets_sent += 1
        self.bytes_sent += len(data)
        if len(data) < self.min_packet_size or self.min_packet_size == 0:
            self.min_packet_size = len(data)
        if len(data) > self.max_packet_size:
            self.max_packet_size = len(data)

    def recv(self, *, data: bytes) -> None:
        self.packets_recv += 1
        self.bytes_recv += len(data)
        if len(data) < self.min_packet_size or self.min_packet_size == 0:
            self.min_packet_size = len(data)
        if len(data) > self.max_packet_size:
            self.max_packet_size = len(data)

    def processing_error(self) -> None:
        self.processing_errors += 1

    def token_error(self) -> None:
        self.token_errors += 1

    def type_error(self) -> None:
        self.type_errors += 1

    def timeout_error(self) -> None:
        self.timeout_errors += 1

    def connection_error(self) -> None:
        self.connection_errors += 1

    def serving_error(self) -> None:
        self.serving_errors += 1

    def resending_error(self) -> None:
        self.resending_errors += 1

    def send_buf_overload(self) -> None:
        self.send_buf_overloads += 1

    def recv_buf_overload(self) -> None:
        self.recv_buf_overloads += 1



import asyncio
from typing import Tuple
import logging

from .packet import Packet
from .telemetry import Telemetry
from .tuning import Tuning
from .ouija import Ouija

from typing import TYPE_CHECKING
if TYPE_CHECKING:   # pragma: no cover
    from proxy import Proxy


logging.basicConfig(
    format='%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d:%H:%M:%S',
    level=logging.ERROR,
)
logger = logging.getLogger(__name__)


class Link(Ouija):
    proxy: 'Proxy'
    addr: Tuple[str, int]

    def __init__(self,  *, telemetry: Telemetry,  proxy: 'Proxy', addr: Tuple[str, int], tuning: Tuning) -> None:
        self.telemetry = telemetry
        self.proxy = proxy
        self.addr = addr
        self.tuning = tuning
        self.reader = None
        self.writer = None
        self.remote_host = None
        self.remote_port = None
        self.opened = asyncio.Event()
        self.sent_buf = dict()
        self.sent_seq = 0
        self.read_closed = asyncio.Event()
        self.recv_buf = dict()
        self.recv_seq = 0
        self.write_closed = asyncio.Event()

    async def sendto(self, *, data: bytes) -> None:
        self.proxy.transport.sendto(data, self.addr)

    async def on_open(self, *, packet: Packet) -> bool:
        if self.opened.is_set():
            await self.send_ack_open()
            return False

        self.remote_host = packet.host
        self.remote_port = packet.port
        self.reader, self.writer = await asyncio.open_connection(self.remote_host, self.remote_port)
        self.opened.set()
        asyncio.create_task(self.serve())
        self.proxy.links[self.addr] = self
        await self.send_ack_open()
        return True

    async def on_serve(self) -> bool:
        return True     # pragma: no cover

    async def on_close(self) -> None:
        self.proxy.links.pop(self.addr, None)



from ouija import RawParser


def test_rawparser_connect():
    request = RawParser(data=b'CONNECT example.com:443 HTTP/1.1')
    assert request.method == 'CONNECT'
    assert request.uri == 'example.com:443'
    assert request.host == 'example.com'
    assert request.port == 443
    assert not request.error


def test_rawparser_get():
    request = RawParser(data=b'GET example.com HTTP/1.1')
    assert not request.error


def test_rawparser_error():
    request = RawParser(data=b'GET /index.html HTTP/1.1')
    assert request.error


def test_rawparser():
    request = RawParser(data=b'CONNECT example.com:443 HTTP/1.1')
    expected = dict(URI=request.uri, HOST=request.host, PORT=request.port, METHOD=request.method)
    assert str(request) == str(expected)



import asyncio
import logging
import time
from typing import Optional, Dict

from .telemetry import Telemetry
from .tuning import Tuning
from .packet import Phase, Packet, Sent, Received


logging.basicConfig(
    format='%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d:%H:%M:%S',
    level=logging.ERROR,
)
logger = logging.getLogger(__name__)


class Ouija:
    telemetry: Telemetry
    tuning: Tuning
    reader: Optional[asyncio.StreamReader]
    writer: Optional[asyncio.StreamWriter]
    remote_host: Optional[str]
    remote_port: Optional[int]
    opened: asyncio.Event
    sent_buf: Dict[int, Sent]
    sent_seq: int
    read_closed: asyncio.Event
    recv_buf: Dict[int, Received]
    recv_seq: int
    write_closed: asyncio.Event

    async def sendto(self, *, data: bytes) -> None:
        raise NotImplementedError

    async def send(self, *, data: bytes) -> None:
        await self.sendto(data=data)
        self.telemetry.send(data=data)

    async def send_packet(self, *, packet: Packet) -> bytes:
        data = packet.binary(fernet=self.tuning.fernet)
        await self.send(data=data)
        return data

    async def send_retry(self, *, packet: Packet, event: asyncio.Event) -> bool:
        for _ in range(self.tuning.udp_retries):
            await self.send_packet(packet=packet)
            try:
                await asyncio.wait_for(event.wait(), self.tuning.udp_timeout)
            except asyncio.TimeoutError:
                continue
            else:
                return True
        else:
            return False

    async def read(self) -> bytes:
        return await self.reader.read(self.tuning.tcp_buffer)

    async def write(self, *, data: bytes, drain: bool) -> None:
        self.writer.write(data)
        if drain:
            await self.writer.drain()

    async def recv(self, *, seq: int, data: bytes, drain: bool) -> None:
        if seq >= self.recv_seq:
            self.recv_buf[seq] = Received(data=data, drain=drain)

        for seq in sorted(self.recv_buf.keys()):
            if seq != self.recv_seq:
                continue

            recv = self.recv_buf.pop(seq)
            await self.write(data=recv.data, drain=recv.drain)
            self.recv_seq += 1

        await self.send_ack_data(seq=seq)

        if len(self.recv_buf) >= self.tuning.udp_capacity:
            await self.close()
            self.telemetry.recv_buf_overload()

    async def enqueue_send(self, *, data: bytes, drain: bool) -> None:
        data_packet = Packet(
            phase=Phase.DATA,
            ack=False,
            seq=self.sent_seq,
            data=data,
            drain=drain,
        )
        self.sent_buf[self.sent_seq] = Sent(data=await self.send_packet(packet=data_packet))
        self.sent_seq += 1

        if len(self.sent_buf) >= self.tuning.udp_capacity:
            await self.close()
            self.telemetry.send_buf_overload()

    async def dequeue_send(self, *, seq: int) -> None:
        self.sent_buf.pop(seq, None)

    async def send_open(self) -> bool:
        open_packet = Packet(
            phase=Phase.OPEN,
            ack=False,
            token=self.tuning.token,
            host=self.remote_host,
            port=self.remote_port,
        )
        if not await self.send_retry(packet=open_packet, event=self.opened):
            return False

        return True

    async def send_ack_open(self) -> None:
        open_ack_packet = Packet(
            phase=Phase.OPEN,
            ack=True,
            token=self.tuning.token,
        )
        await self.send_packet(packet=open_ack_packet)

    async def check_token(self, *, token: str) -> bool:
        if token == self.tuning.token:
            return True

        await self.close()
        self.telemetry.token_error()
        return False

    async def send_ack_data(self, *, seq: int) -> None:
        data_ack_packet = Packet(
            phase=Phase.DATA,
            ack=True,
            seq=seq,
        )
        await self.send_packet(packet=data_ack_packet)

    async def send_close(self) -> None:
        close_packet = Packet(
            phase=Phase.CLOSE,
            ack=False,
        )
        await self.send_retry(packet=close_packet, event=self.read_closed)

    async def send_ack_close(self) -> None:
        close_ack_packet = Packet(
            phase=Phase.CLOSE,
            ack=True,
        )
        await self.send_packet(packet=close_ack_packet)

    async def on_open(self, packet: Packet) -> bool:
        raise NotImplementedError

    async def process_packet(self, *, data: bytes) -> None:
        self.telemetry.recv(data=data)
        packet = Packet.packet(data=data, fernet=self.tuning.fernet)

        match packet.phase:
            case Phase.OPEN:
                if not await self.check_token(token=packet.token):
                    return

                if not await self.on_open(packet=packet):
                    return

                self.telemetry.open()
            case Phase.DATA:
                if not self.opened.is_set() or self.write_closed.is_set():
                    return

                if packet.ack:
                    await self.dequeue_send(seq=packet.seq)
                else:
                    await self.recv(seq=packet.seq, data=packet.data, drain=packet.drain)
            case Phase.CLOSE:
                if packet.ack:
                    self.read_closed.set()
                else:
                    await self.send_ack_close()
                    self.write_closed.set()
            case _:     # pragma: no cover
                self.telemetry.type_error()

    async def process(self, *, data: bytes) -> None:
        try:
            await self.process_packet(data=data)
        except ConnectionError as e:
            logger.error(e)
            self.telemetry.connection_error()
        except Exception as e:
            await self.close()
            logger.error(e)
            self.telemetry.processing_error()

    async def resend_packets(self) -> None:
        while self.opened.is_set() or self.sent_buf:
            await asyncio.sleep(self.tuning.udp_timeout)

            for seq in sorted(self.sent_buf.keys()):
                delta = int(time.time()) - self.sent_buf[seq].timestamp

                if delta >= self.tuning.serving_timeout or self.sent_buf[seq].retries >= self.tuning.udp_retries:
                    await self.close()
                    return

                if delta >= self.tuning.udp_timeout:
                    await self.send(data=self.sent_buf[seq].data)
                    self.sent_buf[seq].retries += 1

        await self.send_close()

        try:
            await asyncio.wait_for(self.write_closed.wait(), self.tuning.serving_timeout)
        except asyncio.TimeoutError:
            pass

    async def resend(self) -> None:
        try:
            await asyncio.wait_for(self.resend_packets(), self.tuning.serving_timeout)
        except asyncio.TimeoutError:
            self.telemetry.timeout_error()
        except Exception as e:
            logger.error(e)
            self.telemetry.resending_error()
        finally:
            await self.close()

    async def on_serve(self) -> bool:
        raise NotImplementedError

    async def serve_stream(self) -> None:
        if not await self.on_serve():
            return

        asyncio.create_task(self.resend())

        while self.opened.is_set():
            try:
                data = await asyncio.wait_for(self.read(), self.tuning.tcp_timeout)
            except TimeoutError:
                continue

            if not data:
                break

            for idx in range(0, len(data), self.tuning.udp_payload):
                await self.enqueue_send(
                    data=data[idx:idx + self.tuning.udp_payload],
                    drain=True if len(data) - idx <= self.tuning.udp_payload else False,
                )

    async def serve(self) -> None:
        try:
            await asyncio.wait_for(self.serve_stream(), self.tuning.serving_timeout)
        except asyncio.TimeoutError:
            self.telemetry.timeout_error()
        except ConnectionError as e:
            logger.error(e)
            self.telemetry.connection_error()
        except Exception as e:
            logger.error(e)
            self.telemetry.serving_error()

    async def on_close(self) -> None:
        raise NotImplementedError

    async def close(self) -> None:
        if self.opened.is_set():
            self.telemetry.close()

        self.opened.clear()
        self.read_closed.set()
        self.write_closed.set()

        if isinstance(self.writer, asyncio.StreamWriter) and not self.writer.is_closing():
            self.writer.close()
            await self.writer.wait_closed()

        await self.on_close()



import asyncio
from unittest.mock import AsyncMock

import pytest

from ouija import Packet, Phase


@pytest.mark.asyncio
@pytest.mark.xfail(raises=NotImplementedError)
async def test_sendto(ouija_test, data_test):
    await ouija_test.sendto(data=data_test)


@pytest.mark.asyncio
async def test_send(ouija_test, data_test):
    ouija_test.sendto = AsyncMock()
    await ouija_test.send(data=data_test)
    ouija_test.sendto.assert_awaited_with(data=data_test)


@pytest.mark.asyncio
async def test_send_packet(ouija_test, data_test):
    ouija_test.send = AsyncMock()
    packet = Packet(phase=Phase.DATA, ack=False, seq=0, data=data_test, drain=True)
    await ouija_test.send_packet(packet=packet)
    ouija_test.send.assert_awaited()


@pytest.mark.asyncio
async def test_send_retry(ouija_test, data_test):
    ouija_test.send_packet = AsyncMock()
    packet = Packet(phase=Phase.DATA, ack=False, seq=0, data=data_test, drain=True)
    event = asyncio.Event()
    event.set()
    result = await ouija_test.send_retry(packet=packet, event=event)
    ouija_test.send_packet.assert_awaited()
    assert result


@pytest.mark.asyncio
async def test_send_retry_event_not_set(ouija_test, data_test):
    ouija_test.send_packet = AsyncMock()
    packet = Packet(phase=Phase.DATA, ack=False, seq=0, data=data_test, drain=True)
    event = asyncio.Event()
    result = await ouija_test.send_retry(packet=packet, event=event)
    ouija_test.send_packet.assert_awaited()
    assert not result


@pytest.mark.asyncio
async def test_read(ouija_test):
    ouija_test.reader = AsyncMock()
    await ouija_test.read()
    ouija_test.reader.read.assert_called()


@pytest.mark.asyncio
async def test_write(ouija_test, data_test):
    ouija_test.writer = AsyncMock()
    await ouija_test.write(data=data_test, drain=False)
    ouija_test.writer.write.assert_called()


@pytest.mark.asyncio
async def test_write_drain(ouija_test, data_test):
    ouija_test.writer = AsyncMock()
    await ouija_test.write(data=data_test, drain=True)
    ouija_test.writer.write.assert_called()
    ouija_test.writer.drain.assert_awaited()


@pytest.mark.asyncio
async def test_recv(ouija_test, data_test):
    ouija_test.write = AsyncMock()
    ouija_test.send_ack_data = AsyncMock()
    await ouija_test.recv(seq=1, data=data_test, drain=False)
    await ouija_test.recv(seq=2, data=data_test, drain=False)
    await ouija_test.recv(seq=0, data=data_test, drain=False)
    ouija_test.write.assert_awaited()
    ouija_test.send_ack_data.assert_awaited()


@pytest.mark.asyncio
async def test_recv_overload(ouija_test, data_test):
    ouija_test.write = AsyncMock()
    ouija_test.send_ack_data = AsyncMock()
    ouija_test.close = AsyncMock()
    ouija_test.tuning.udp_capacity = 1
    await ouija_test.recv(seq=1, data=data_test, drain=False)
    await ouija_test.recv(seq=2, data=data_test, drain=False)
    ouija_test.close.assert_awaited()


@pytest.mark.asyncio
async def test_enqueue_send(ouija_test, data_test):
    ouija_test.send_packet = AsyncMock()
    await ouija_test.enqueue_send(data=data_test, drain=True)
    ouija_test.send_packet.assert_awaited()


@pytest.mark.asyncio
async def test_enqueue_send_overload(ouija_test, data_test):
    ouija_test.send_packet = AsyncMock()
    ouija_test.close = AsyncMock()
    ouija_test.tuning.udp_capacity = 1
    await ouija_test.enqueue_send(data=data_test, drain=False)
    await ouija_test.enqueue_send(data=data_test, drain=True)
    ouija_test.send_packet.assert_awaited()
    ouija_test.close.assert_awaited()


@pytest.mark.asyncio
async def test_dequeue_send(ouija_test):
    ouija_test.sent_buf = AsyncMock()
    await ouija_test.dequeue_send(seq=1)
    ouija_test.sent_buf.pop.assert_called()


@pytest.mark.asyncio
async def test_send_open(ouija_test):
    ouija_test.send_retry = AsyncMock(return_value=True)
    assert await ouija_test.send_open()
    ouija_test.send_retry.assert_awaited()


@pytest.mark.asyncio
async def test_send_open_send_retry_false(ouija_test):
    ouija_test.send_retry = AsyncMock(return_value=False)
    assert not await ouija_test.send_open()
    ouija_test.send_retry.assert_awaited()


@pytest.mark.asyncio
async def test_send_ack_open(ouija_test):
    ouija_test.send_packet = AsyncMock()
    await ouija_test.send_ack_open()
    ouija_test.send_packet.assert_awaited()


@pytest.mark.asyncio
async def test_check_token(ouija_test, token_test):
    assert await ouija_test.check_token(token=token_test)


@pytest.mark.asyncio
async def test_check_token_false(ouija_test, token_test):
    ouija_test.close = AsyncMock()
    result = await ouija_test.check_token(token='token')
    ouija_test.close.assert_awaited()
    assert not result


@pytest.mark.asyncio
async def test_send_ack_data(ouija_test):
    ouija_test.send_packet = AsyncMock()
    await ouija_test.send_ack_data(seq=0)
    ouija_test.send_packet.assert_awaited()


@pytest.mark.asyncio
async def test_send_close(ouija_test):
    ouija_test.send_retry = AsyncMock()
    await ouija_test.send_close()
    ouija_test.send_retry.assert_awaited()


@pytest.mark.asyncio
@pytest.mark.xfail(raises=NotImplementedError)
async def test_on_open(ouija_test, token_test):
    packet = Packet(
        phase=Phase.OPEN,
        ack=False,
        token=token_test,
        host='example.com',
        port=443,
    )
    await ouija_test.on_open(packet=packet)


@pytest.mark.asyncio
async def test_send_ack_close(ouija_test):
    ouija_test.send_packet = AsyncMock()
    await ouija_test.send_ack_close()
    ouija_test.send_packet.assert_awaited()


@pytest.mark.asyncio
async def test_process_packet_open(ouija_test, token_test):
    ouija_test.check_token = AsyncMock(return_value=True)
    ouija_test.on_open = AsyncMock(return_value=True)
    packet = Packet(
        phase=Phase.OPEN,
        ack=False,
        token=token_test,
        host='example.com',
        port=443,
    )
    await ouija_test.process_packet(data=packet.binary(fernet=ouija_test.tuning.fernet))
    ouija_test.check_token.assert_awaited()
    ouija_test.on_open.assert_awaited()


@pytest.mark.asyncio
async def test_process_packet_open_check_token_false(ouija_test, token_test):
    ouija_test.check_token = AsyncMock(return_value=False)
    packet = Packet(
        phase=Phase.OPEN,
        ack=False,
        token=token_test,
        host='example.com',
        port=443,
    )
    await ouija_test.process_packet(data=packet.binary(fernet=ouija_test.tuning.fernet))
    ouija_test.check_token.assert_awaited()


@pytest.mark.asyncio
async def test_process_packet_open_on_open_false(ouija_test, token_test):
    ouija_test.on_open = AsyncMock(return_value=False)
    packet = Packet(
        phase=Phase.OPEN,
        ack=False,
        token=token_test,
        host='example.com',
        port=443,
    )
    await ouija_test.process_packet(data=packet.binary(fernet=ouija_test.tuning.fernet))
    ouija_test.on_open.assert_awaited()


@pytest.mark.asyncio
async def test_process_packet_data_ack(ouija_test):
    ouija_test.opened.set()
    ouija_test.dequeue_send = AsyncMock()
    packet = Packet(
        phase=Phase.DATA,
        ack=True,
        seq=0,
    )
    await ouija_test.process_packet(data=packet.binary(fernet=ouija_test.tuning.fernet))
    ouija_test.dequeue_send.assert_awaited()


@pytest.mark.asyncio
async def test_process_packet_data(ouija_test, data_test):
    ouija_test.opened.set()
    ouija_test.recv = AsyncMock()
    packet = Packet(
        phase=Phase.DATA,
        ack=False,
        seq=0,
        data=data_test,
        drain=False,
    )
    await ouija_test.process_packet(data=packet.binary(fernet=ouija_test.tuning.fernet))
    ouija_test.recv.assert_awaited()


@pytest.mark.asyncio
async def test_process_packet_data_not_opened(ouija_test, data_test):
    ouija_test.dequeue_send = AsyncMock()
    ouija_test.recv = AsyncMock()
    packet = Packet(
        phase=Phase.DATA,
        ack=False,
        seq=0,
        data=data_test,
        drain=False,
    )
    await ouija_test.process_packet(data=packet.binary(fernet=ouija_test.tuning.fernet))
    ouija_test.dequeue_send.assert_not_awaited()
    ouija_test.recv.assert_not_awaited()


@pytest.mark.asyncio
async def test_process_packet_close_ack(ouija_test):
    packet = Packet(
        phase=Phase.CLOSE,
        ack=True,
    )
    await ouija_test.process_packet(data=packet.binary(fernet=ouija_test.tuning.fernet))
    assert ouija_test.read_closed.is_set()


@pytest.mark.asyncio
async def test_process_packet_close(ouija_test):
    ouija_test.send_ack_close = AsyncMock()
    packet = Packet(
        phase=Phase.CLOSE,
        ack=False,
    )
    await ouija_test.process_packet(data=packet.binary(fernet=ouija_test.tuning.fernet))
    ouija_test.send_ack_close.assert_awaited()


@pytest.mark.asyncio
async def test_process(ouija_test):
    ouija_test.process_packet = AsyncMock()
    packet = Packet(
        phase=Phase.CLOSE,
        ack=False,
    )
    await ouija_test.process(data=packet.binary(fernet=ouija_test.tuning.fernet))
    ouija_test.process_packet.assert_awaited()


@pytest.mark.asyncio
@pytest.mark.xfail(raises=ConnectionError)
async def test_process_connectionerror(ouija_test):
    ouija_test.process_packet = AsyncMock()
    ouija_test.process_packet.side_effect = ConnectionError()
    packet = Packet(
        phase=Phase.CLOSE,
        ack=False,
    )
    await ouija_test.process(data=packet.binary(fernet=ouija_test.tuning.fernet))
    ouija_test.process_packet.assert_awaited()


@pytest.mark.asyncio
@pytest.mark.xfail(raises=Exception)
async def test_process_exception(ouija_test):
    ouija_test.process_packet = AsyncMock()
    ouija_test.process_packet.side_effect = Exception()
    ouija_test.close = AsyncMock()
    packet = Packet(
        phase=Phase.CLOSE,
        ack=False,
    )
    await ouija_test.process(data=packet.binary(fernet=ouija_test.tuning.fernet))
    ouija_test.process_packet.assert_awaited()
    ouija_test.close.assert_awaited()


@pytest.mark.asyncio
async def test_resend_packets(ouija_test, data_test):
    ouija_test.close = AsyncMock()
    ouija_test.send = AsyncMock()
    ouija_test.send_close = AsyncMock()
    await ouija_test.enqueue_send(data=data_test, drain=True)
    await ouija_test.resend_packets()
    ouija_test.close.assert_awaited()
    ouija_test.send.assert_awaited()
    ouija_test.send_close.assert_not_awaited()


@pytest.mark.asyncio
async def test_resend_packets_empty(ouija_test):
    ouija_test.send_close = AsyncMock()
    await ouija_test.resend_packets()
    ouija_test.send_close.assert_awaited()


@pytest.mark.asyncio
async def test_resend(ouija_test):
    ouija_test.resend_packets = AsyncMock()
    ouija_test.close = AsyncMock()
    await ouija_test.resend()
    ouija_test.resend_packets.assert_awaited()
    ouija_test.close.assert_awaited()


@pytest.mark.asyncio
@pytest.mark.xfail(raises=TimeoutError)
async def test_resend_timeouterror(ouija_test):
    ouija_test.resend_packets = AsyncMock()
    ouija_test.resend_packets.side_effect = TimeoutError()
    ouija_test.close = AsyncMock()
    await ouija_test.resend()
    ouija_test.resend_packets.assert_awaited()
    ouija_test.close.assert_awaited()


@pytest.mark.asyncio
@pytest.mark.xfail(raises=Exception)
async def test_resend_exception(ouija_test):
    ouija_test.resend_packets = AsyncMock()
    ouija_test.resend_packets.side_effect = Exception()
    await ouija_test.resend()
    ouija_test.resend_packets.assert_awaited()
    ouija_test.close.assert_awaited()


@pytest.mark.asyncio
@pytest.mark.xfail(raises=NotImplementedError)
async def test_on_serve(ouija_test):
    await ouija_test.on_serve()


@pytest.mark.asyncio
async def test_serve_stream(ouija_test, data_test):
    async def resetter():
        await asyncio.sleep(3)
        ouija_test.opened.clear()
    ouija_test.on_serve = AsyncMock(return_value=True)
    ouija_test.resend = AsyncMock()
    ouija_test.read = AsyncMock(return_value=data_test)
    ouija_test.enqueue_send = AsyncMock()
    ouija_test.opened.set()
    asyncio.create_task(resetter())
    await ouija_test.serve_stream()
    ouija_test.on_serve.assert_awaited()
    ouija_test.resend.assert_awaited()
    ouija_test.read.assert_awaited()
    ouija_test.enqueue_send.assert_awaited()


@pytest.mark.asyncio
async def test_serve_stream_on_serve_false(ouija_test, data_test):
    async def resetter():
        await asyncio.sleep(3)
        ouija_test.opened.clear()
    ouija_test.on_serve = AsyncMock(return_value=False)
    ouija_test.resend = AsyncMock()
    ouija_test.read = AsyncMock(return_value=data_test)
    ouija_test.enqueue_send = AsyncMock()
    ouija_test.opened.set()
    asyncio.create_task(resetter())
    await ouija_test.serve_stream()
    ouija_test.on_serve.assert_awaited()
    ouija_test.resend.assert_not_awaited()
    ouija_test.read.assert_not_awaited()
    ouija_test.enqueue_send.assert_not_awaited()


@pytest.mark.asyncio
@pytest.mark.xfail(raises=TimeoutError)
async def test_serve_stream_timeouterror(ouija_test, data_test):
    async def resetter():
        await asyncio.sleep(3)
        ouija_test.opened.clear()
    ouija_test.on_serve = AsyncMock(return_value=True)
    ouija_test.resend = AsyncMock()
    ouija_test.read = AsyncMock(return_value=data_test)
    ouija_test.read.side_effect = TimeoutError()
    ouija_test.enqueue_send = AsyncMock()
    ouija_test.opened.set()
    asyncio.create_task(resetter())
    await ouija_test.serve_stream()
    ouija_test.on_serve.assert_awaited()
    ouija_test.resend.assert_awaited()
    ouija_test.read.assert_awaited()
    ouija_test.enqueue_send.assert_not_awaited()


@pytest.mark.asyncio
async def test_serve_stream_empty(ouija_test, data_test):
    async def resetter():
        await asyncio.sleep(3)
        ouija_test.opened.clear()
    ouija_test.on_serve = AsyncMock(return_value=True)
    ouija_test.resend = AsyncMock()
    ouija_test.read = AsyncMock(return_value=b'')
    ouija_test.enqueue_send = AsyncMock()
    ouija_test.opened.set()
    asyncio.create_task(resetter())
    await ouija_test.serve_stream()
    ouija_test.on_serve.assert_awaited()
    ouija_test.resend.assert_awaited()
    ouija_test.read.assert_awaited()
    ouija_test.enqueue_send.assert_not_awaited()


@pytest.mark.asyncio
async def test_serve(ouija_test):
    ouija_test.serve_stream = AsyncMock()
    await ouija_test.serve()
    ouija_test.serve_stream.assert_awaited()


@pytest.mark.asyncio
@pytest.mark.xfail(raises=TimeoutError)
async def test_serve_timeouterror(ouija_test):
    ouija_test.serve_stream = AsyncMock()
    ouija_test.serve_stream.side_effect = TimeoutError()
    await ouija_test.serve()
    ouija_test.serve_stream.assert_awaited()


@pytest.mark.asyncio
@pytest.mark.xfail(raises=ConnectionError)
async def test_serve_connectionerror(ouija_test):
    ouija_test.serve_stream = AsyncMock()
    ouija_test.serve_stream.side_effect = ConnectionError()
    await ouija_test.serve()
    ouija_test.serve_stream.assert_awaited()


@pytest.mark.asyncio
@pytest.mark.xfail(raises=Exception)
async def test_serve_exception(ouija_test):
    ouija_test.serve_stream = AsyncMock()
    ouija_test.serve_stream.side_effect = Exception()
    await ouija_test.serve()
    ouija_test.serve_stream.assert_awaited()


@pytest.mark.asyncio
async def test_close(ouija_test):
    ouija_test.opened.set()
    ouija_test.writer = AsyncMock(spec=asyncio.StreamWriter)
    ouija_test.writer.is_closing = lambda: False
    ouija_test.on_close = AsyncMock()
    await ouija_test.close()
    assert not ouija_test.opened.is_set()
    assert ouija_test.read_closed.is_set()
    assert ouija_test.write_closed.is_set()
    ouija_test.writer.close.assert_called()
    ouija_test.writer.wait_closed.assert_awaited()
    ouija_test.on_close.assert_awaited()



import dataclasses

from cryptography.fernet import Fernet


@dataclasses.dataclass(kw_only=True)
class Tuning:
    fernet: Fernet
    token: str
    serving_timeout: int
    tcp_buffer: int
    tcp_timeout: int
    udp_payload: int
    udp_timeout: int
    udp_retries: int
    udp_capacity: int



from unittest.mock import Mock, AsyncMock

import pytest

from ouija import Packet, Phase


def test_relay_connection_made(relay_test):
    mock = Mock()
    relay_test.connection_made(mock)
    assert relay_test.transport == mock


@pytest.mark.asyncio
async def test_relay_datagram_received_async(relay_test, data_test):
    relay_test.process = AsyncMock()
    await relay_test.datagram_received_async(data=data_test, addr='127.0.0.1')
    relay_test.process.assert_awaited()


@pytest.mark.asyncio
async def test_relay_datagram_received(relay_test, data_test):
    relay_test.datagram_received_async = AsyncMock()
    relay_test.datagram_received(data_test, '127.0.0.1')
    relay_test.datagram_received_async.assert_called()


@pytest.mark.asyncio
async def test_relay_connection_lost(relay_test):
    relay_test.close = AsyncMock()
    relay_test.connection_lost(Exception())
    relay_test.close.assert_called()


@pytest.mark.asyncio
async def test_relay_sendto(relay_test, data_test):
    relay_test.transport = Mock()
    await relay_test.sendto(data=data_test)
    relay_test.transport.sendto.assert_called()


@pytest.mark.asyncio
async def test_relay_on_open(relay_test, token_test):
    relay_test.write = AsyncMock()
    packet = Packet(
        phase=Phase.OPEN,
        ack=True,
        token=token_test,
    )
    assert await relay_test.on_open(packet=packet)
    relay_test.write.assert_awaited()
    assert relay_test.opened.is_set()


@pytest.mark.asyncio
async def test_relay_on_open_not_ack(relay_test, token_test):
    packet = Packet(
        phase=Phase.OPEN,
        ack=False,
        token=token_test,
    )
    assert not await relay_test.on_open(packet=packet)
    assert not relay_test.opened.is_set()


@pytest.mark.asyncio
async def test_relay_on_open_opened(relay_test, token_test):
    relay_test.opened.set()
    packet = Packet(
        phase=Phase.OPEN,
        ack=True,
        token=token_test,
    )
    assert not await relay_test.on_open(packet=packet)


@pytest.mark.asyncio
async def test_relay_on_serve(relay_test, token_test):
    relay_test.send_open = AsyncMock(return_value=True)
    assert await relay_test.on_serve()
    relay_test.send_open.assert_awaited()


@pytest.mark.asyncio
async def test_relay_on_serve_send_open_false(relay_test, token_test):
    relay_test.send_open = AsyncMock(return_value=False)
    assert not await relay_test.on_serve()
    relay_test.send_open.assert_awaited()


@pytest.mark.asyncio
async def test_relay_on_close(relay_test, token_test):
    relay_test.transport = AsyncMock()
    relay_test.transport.is_closing = lambda: False
    await relay_test.on_close()
    relay_test.transport.close.assert_called()


@pytest.mark.asyncio
async def test_relay_on_close_closing(relay_test, token_test):
    relay_test.transport = AsyncMock()
    relay_test.transport.is_closing = lambda: True
    await relay_test.on_close()
    relay_test.transport.close.assert_not_called()
